// -*- Mode: vala; indent-tabs-mode: nil; tab-width: 4 -*-
/*-
 * Copyright (c) 2014 Pantheon Online Accounts Developers (http://launchpad.net/pantheon-online-accounts)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authored by: Corentin Noël <corentin@elementaryos.org>
 */

public class E.Outlook2BackendFactory : E.CollectionBackendFactory {

    /* IMAP Configuration Details */
    private const string OUTLOOK_IMAP_BACKEND_NAME = "imapx";
    private const string OUTLOOK_IMAP_HOST ="imap-mail.outlook.com";
    private const uint16 OUTLOOK_IMAP_PORT = 993;
    private const Camel.NetworkSecurityMethod OUTLOOK_IMAP_SECURITY_METHOD = Camel.NetworkSecurityMethod.SSL_ON_ALTERNATE_PORT;

    /* SMTP Configuration Details */
    private const string OUTLOOK_SMTP_BACKEND_NAME = "smtp";
    private const string OUTLOOK_SMTP_HOST = "smtp-mail.outlook.com";
    private const uint16 OUTLOOK_SMTP_PORT = 587;
    private const Camel.NetworkSecurityMethod OUTLOOK_SMTP_SECURITY_METHOD = Camel.NetworkSecurityMethod.STARTTLS_ON_STANDARD_PORT;

    static construct {
        factory_name = "outlook2";
        backend_type = typeof (Outlook2Backend);
    }

    public override void prepare_mail (E.Source mail_account_source, E.Source mail_identity_source, E.Source mail_transport_source) {
        warning ("HEY3");
        base.prepare_mail (mail_account_source, mail_identity_source, mail_transport_source);
        prepare_mail_account_source (mail_account_source);
        prepare_mail_transport_source (mail_transport_source);
    }

    private void prepare_mail_account_source (E.Source source) {
        var extension = (E.SourceBackend) source.get_extension (E.SOURCE_EXTENSION_MAIL_ACCOUNT);
        extension.set_backend_name (OUTLOOK_IMAP_BACKEND_NAME);
        var extension_name = E.SourceCamel.get_extension_name (OUTLOOK_IMAP_BACKEND_NAME);
        var camel_extension = (E.SourceCamel) source.get_extension (extension_name);
        var settings = (Camel.NetworkSettings) camel_extension.get_settings ();

        var auth_extension = (E.SourceAuthentication)source.get_extension (E.SOURCE_EXTENSION_AUTHENTICATION);
        auth_extension.set_method ("OAuth2");

        /* The "auth-mechanism" should be determined elsewhere. */
        settings.host = OUTLOOK_IMAP_HOST;
        settings.port = OUTLOOK_IMAP_PORT;
        settings.security_method = OUTLOOK_IMAP_SECURITY_METHOD;
    }

    private void prepare_mail_transport_source (E.Source source) {
        var extension = (E.SourceBackend) source.get_extension (E.SOURCE_EXTENSION_MAIL_TRANSPORT);
        extension.set_backend_name (OUTLOOK_SMTP_BACKEND_NAME);
        var extension_name = E.SourceCamel.get_extension_name (OUTLOOK_SMTP_BACKEND_NAME);
        var camel_extension = (E.SourceCamel) source.get_extension (extension_name);
        var settings = (Camel.NetworkSettings) camel_extension.get_settings ();

        var auth_extension = (E.SourceAuthentication)source.get_extension (E.SOURCE_EXTENSION_AUTHENTICATION);
        auth_extension.set_method ("OAuth2");

        /* The "auth-mechanism" should be determined elsewhere. */
        settings.host = OUTLOOK_SMTP_HOST;
        settings.port = OUTLOOK_SMTP_PORT;
        settings.security_method = OUTLOOK_SMTP_SECURITY_METHOD;
    }
}

public class E.Outlook2Backend : E.CollectionBackend {
    /* Calendar Configuration Details */
    private const string OUTLOOK_CALENDAR_BACKEND_NAME = "outlook2";
    private const string OUTLOOK_CALENDAR_RESOURCE_ID = "Calendar";

    /* Contacts Configuration Details */
    private const string OUTLOOK_CONTACTS_BACKEND_NAME = "outlook2";
    private const string OUTLOOK_CONTACTS_RESOURCE_ID = "Contacts";

    private static void soup_authenticate (Soup.Message msg, Soup.Auth auth, bool retrying, E.Source src) {
        string access_token = "";
        int expires_in = 0;
        try {
            src.get_oauth2_access_token_sync (null, out access_token, out expires_in);
            ((E.SoupAuthBearer) auth).set_access_token (access_token, expires_in);
        } catch (Error e) {
            critical (e.message);
        }
    }

    //TODO: check for new/removed calendars
    public override void populate () {
        base.populate ();

        if (list_calendar_sources () == null)
            add_calendars ();

        if (list_contacts_sources () == null)
            add_contacts ();
    }

    public override void child_added (E.ServerSideSource child_source) {
        var child_auth_extension = (E.SourceAuthentication) child_source.get_extension (E.SOURCE_EXTENSION_AUTHENTICATION);
        child_auth_extension.set_method ("OAuth2");
        var collection_source = get_source ();
        var collection_extension = (E.SourceCollection) collection_source.get_extension (E.SOURCE_EXTENSION_COLLECTION);
        collection_extension.bind_property ("identity", child_auth_extension, "user", GLib.BindingFlags.SYNC_CREATE);

        base.child_added (child_source);
    
    }

    private void add_calendars () {
        var collection_source = get_source ();
        var source = new_child (OUTLOOK_CALENDAR_RESOURCE_ID);
        source.set_display_name (_("Calendar"));
        var collection_extension = collection_source.get_extension (E.SOURCE_EXTENSION_COLLECTION);

        /* Configure the address book source. */
        var extension = source.get_extension (E.SOURCE_EXTENSION_CALENDAR);
        ((E.SourceBackend) extension).set_backend_name (OUTLOOK_CALENDAR_BACKEND_NAME);
        extension = source.get_extension (E.SOURCE_EXTENSION_AUTHENTICATION);
        collection_extension.bind_property ("identity", extension, "user", GLib.BindingFlags.SYNC_CREATE);

        ref_server ().add_source (source);

        /* Check for all calendars */
        try {
            GLib.List<unowned Json.Node> elements = get_calendars_with_source (source);
            foreach (unowned Json.Node item in elements) {
                if (item.get_node_type () != Json.NodeType.OBJECT)
                    continue;
                var object = item.dup_object ();
                // We can't select a calendar here, let's use the default one
                if (object.get_boolean_member ("is_default") == true) {
                    var id = object.get_string_member ("id");
                    var ressource_extension = (E.SourceResource) source.get_extension (E.SOURCE_EXTENSION_RESOURCE);
                    ressource_extension.set_identity (id);
                    continue;
                } else {
                    create_calendar_from_object (object);
                }
            }
        } catch (Error e) {
            critical (e.message);
        }
    }

    private void create_calendar_from_object (Json.Object object) {
        var collection_source = get_source ();
        var collection_extension = collection_source.get_extension (E.SOURCE_EXTENSION_COLLECTION);
        var id = object.get_string_member ("id");
        var name = object.get_string_member ("name");
        var permission = object.get_string_member ("permissions");
        var new_source = new_child (id);
        new_source.set_display_name (name);

        /* Configure the address book source. */
        var new_extension = new_source.get_extension (E.SOURCE_EXTENSION_CALENDAR);
        ((E.SourceBackend) new_extension).set_backend_name (OUTLOOK_CALENDAR_BACKEND_NAME);
        new_extension = new_source.get_extension (E.SOURCE_EXTENSION_AUTHENTICATION);
        collection_extension.bind_property ("identity", new_extension, "user", GLib.BindingFlags.SYNC_CREATE);

        var ressource_extension = (E.SourceResource) new_source.get_extension (E.SOURCE_EXTENSION_RESOURCE);
        ressource_extension.set_identity (id);

        /*switch (permission) {
            case "read_write":
                ((E.ServerSideSource)new_source).removable = false;
                ((E.ServerSideSource)new_source).writable = true;
                break;
            case "read":
            case "limited_details":
                ((E.ServerSideSource)new_source).removable = false;
                ((E.ServerSideSource)new_source).writable = false;
                break;
            case "free_busy":
                return;
            case "owner":
            case "co_owner":
            default:
                ((E.ServerSideSource)new_source).removable = true;
                ((E.ServerSideSource)new_source).writable = true;
                break;
        }*/

        ref_server ().add_source (new_source);
    }

    private GLib.List<unowned Json.Node> get_calendars_with_source (E.Source source) throws GLib.Error {
        var session = new Soup.Session ();
        session.timeout = 90;
        session.ssl_strict = true;
        session.ssl_use_system_ca_file = true;
        var feature = session.get_feature (typeof (Soup.AuthManager));
        feature.add_feature (typeof (E.SoupAuthBearer));
        session.authenticate.connect ((msg, auth, retrying) => soup_authenticate (msg, auth, retrying, source));
        Soup.Message msg = new Soup.Message ("GET", "https://apis.live.net/v5.0/me/calendars");
        session.send_message (msg);
        var parser = new Json.Parser ();
        try {
            parser.load_from_data ((string)msg.response_body.data);
            var object = parser.get_root ().dup_object ();
            if (object.has_member ("data") == false) {
                throw new GLib.IOError.CONNECTION_REFUSED ("");
            }

            return object.dup_member ("data").get_array ().get_elements ();
        } catch (Error e) {
            throw e;
        }
    }

    private void add_contacts () {
        var collection_source = get_source ();
        var source = new_child (OUTLOOK_CONTACTS_RESOURCE_ID);
        source.set_display_name (_("Contacts"));
        var collection_extension = collection_source.get_extension (E.SOURCE_EXTENSION_COLLECTION);

        /* Configure the address book source. */
        var extension = source.get_extension (E.SOURCE_EXTENSION_ADDRESS_BOOK);
        ((E.SourceBackend) extension).set_backend_name (OUTLOOK_CONTACTS_BACKEND_NAME);

        ref_server ().add_source (source);
    }
}

[ModuleInit]
public static Type e_module_load (GLib.TypeModule type_module) {
    e_outlook_backend_fake_module_load (type_module);
    return typeof (E.Outlook2BackendFactory);
}

[ModuleInit]
public static Type e_outlook_backend_fake_module_load (GLib.TypeModule type_module) {
    return typeof(E.Outlook2Backend);
}

public void e_module_unload (TypeModule type_module) {
    
}