// -*- Mode: vala; indent-tabs-mode: nil; tab-width: 4 -*-
/*-
 * Copyright (c) 2014 Pantheon Online Accounts Developers (http://launchpad.net/pantheon-online-accounts)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authored by: Corentin Noël <corentin@elementaryos.org>
 */

// Main references : http://msdn.microsoft.com/en-us/library/hh243648.aspx
//                   http://msdn.microsoft.com/en-us/library/hh826523.aspx

public class E.CalBackendOutlook : E.CalBackendSync {
    public static string FORMAT = "%FT%T%:z";
    private E.CalBackendStore store;
    private Soup.Session session;
    private bool loaded = false;
    private bool do_offline = false;
    private bool auth_required = true;
    private bool updating_source = false;
    private uint refresh_id = 0;
    private DateTime last_updated;

    construct {
        session = new Soup.Session ();
        session.timeout = 90;
        session.ssl_strict = true;
        session.ssl_use_system_ca_file = true;
        var feature = session.get_feature (typeof (Soup.AuthManager));
        feature.add_feature (typeof (E.SoupAuthBearer));

        session.authenticate.connect ((msg, auth, retrying) => soup_authenticate (msg, auth, retrying));
    }

    /**
     * Converts two datetimes to one TimeType. The first contains the date,
     * its time settings are ignored. The second one contains the time itself.
     */
    public static iCal.TimeType date_time_to_ical (DateTime date, DateTime? time) {

        var result = iCal.TimeType.from_day_of_year
            (date.get_day_of_year (), date.get_year ());
        result.zone = iCal.TimeZone.get_builtin_timezone (E.Cal.util_get_system_timezone_location ());

        if (time != null) {
            result.is_date = 0;
            result.hour = time.get_hour ();
            result.minute = time.get_minute ();
            result.second = time.get_second ();
        } else {
            result.is_date = 1;
            result.hour = 0;
            result.minute = 0;
            result.second = 0;
        }

        return result;
    }

    /**
     * Converts the given TimeType to a DateTime.
     */
    public DateTime ical_to_date_time (iCal.TimeType date) {

        string tzid = date.zone.get_tzid ();
        TimeZone zone = new TimeZone (tzid);

        return new DateTime (zone, date.year, date.month,
            date.day, date.hour, date.minute, date.second);
    }

    public override string get_backend_property (string prop_name) {
        switch (prop_name) {
            case E.CalBackendProperty.CAPABILITIES: 
                return E.CalStaticCapability.NO_THISANDFUTURE + "," + E.CalStaticCapability.NO_THISANDPRIOR + "," + E.CalStaticCapability.REFRESH_SUPPORTED;
            case E.CalBackendProperty.CAL_EMAIL_ADDRESS: 
            case E.CalBackendProperty.ALARM_EMAIL_ADDRESS:
                return null; //TODO: implement it!
            case E.CalBackendProperty.DEFAULT_OBJECT:
                var comp = new E.CalComponent ();
                comp.set_new_vtype (E.CalComponentVType.EVENT);
                return comp.get_as_string ();
        }

        return null;
    }

    public override void shutdown (string test) {
        warning ("");
    }

    public override void open_sync (E.DataCal cal, GLib.Cancellable? cancellable, bool only_if_exists) throws GLib.Error {
        if (get_online () == true) {
            try {
                initialize_backend ();
            } catch (Error e) {
                throw e;
            }
        } else {
            writable = false;
        }
    }

    public override void refresh_sync (E.DataCal cal, GLib.Cancellable? cancellable = null) throws GLib.Error {
        time_to_refresh ();
    }

    public override void start_view (E.DataCalView view) {
        var sexp = view.get_sexp ();
        var sexp_string = sexp.text ();
        bool do_search = true;

        if (sexp_string == "#t") {
            do_search = false;
        }

        time_t occur_start = -1;
        time_t occur_end = -1;
        bool prunning_by_time = sexp.evaluate_occur_times (ref occur_start, ref occur_end);

        SList<E.CalComponent> list;
        if (prunning_by_time == true) {
            list = store.get_components_occuring_in_range (occur_start, occur_end);
        } else {
            list = store.get_components ();
        }

        list.@foreach ((comp) => {
            if (do_search == false || sexp.match_comp (comp, this)) {
                view.notify_components_added_1 (comp);
            }
        });

        view.notify_complete (null);
    }

    public override void create_objects_sync (E.DataCal cal, GLib.Cancellable? cancellable, GLib.SList<string> calobjs, GLib.SList<string> uids, GLib.SList<E.CalComponent> new_components) throws GLib.Error {
        foreach (var calobj in calobjs) {
            var ical = new iCal.Component.from_string (calobj);
            var builder = new Json.Builder ();
            builder.begin_object ();
            builder.set_member_name ("name");
            builder.add_string_value (ical.get_summary ());
            builder.set_member_name ("description");
            builder.add_string_value (ical.get_summary ());
            builder.set_member_name ("start_time");
            var start_datetime = ical_to_date_time (ical.get_dtstart ());
            builder.add_string_value (start_datetime.format (FORMAT));
            builder.set_member_name ("end_time");
            var end_datetime = ical_to_date_time (ical.get_dtend ());
            builder.add_string_value (end_datetime.format (FORMAT));
            builder.set_member_name ("location");
            builder.add_string_value (ical.get_location ());
            /*builder.set_member_name ("is_all_day_event");
            builder.add_boolean_value (false);*/
            builder.end_object ();

            var generator = new Json.Generator ();
            generator.set_root (builder.get_root ());

            var msg = new Soup.Message ("POST", "https://apis.live.net/v5.0/me/events");
            msg.request_headers.append ("Content-Type", "application/json");
            msg.request_body.append_take (generator.to_data (null).data);
            session.send_message (msg);
        }
    }

    public override void modify_objects_sync (E.DataCal cal, GLib.Cancellable? cancellable, GLib.SList<string> calobjs, E.CalObjModType mod, GLib.SList<E.CalComponent> old_components, GLib.SList<E.CalComponent> new_components) throws GLib.Error {
        foreach (var calobj in calobjs) {
            var ical = new iCal.Component.from_string (calobj);
            var builder = new Json.Builder ();
            builder.begin_object ();
            builder.set_member_name ("name");
            builder.add_string_value (ical.get_summary ());
            builder.set_member_name ("description");
            builder.add_string_value (ical.get_summary ());
            builder.set_member_name ("start_time");
            var start_datetime = ical_to_date_time (ical.get_dtstart ());
            builder.add_string_value (start_datetime.format ("%FT%T%:z"));
            builder.set_member_name ("end_time");
            var end_datetime = ical_to_date_time (ical.get_dtend ());
            builder.add_string_value (end_datetime.format ("%FT%T%:z"));
            builder.set_member_name ("location");
            builder.add_string_value (ical.get_location ());
            /*builder.set_member_name ("is_all_day_event");
            builder.add_boolean_value (false);*/
            builder.end_object ();

            var generator = new Json.Generator ();
            generator.set_root (builder.get_root ());

            var msg = new Soup.Message ("PUT", "https://apis.live.net/v5.0/%s".printf (ical.get_relcalid ()));
            msg.request_headers.append ("Content-Type", "application/json");
            msg.request_body.append_take (generator.to_data (null).data);
            session.send_message (msg);
        }
    }

    public override void remove_objects_sync (E.DataCal cal, GLib.Cancellable? cancellable, GLib.SList<unowned E.CalComponentId?> ids, E.CalObjModType mod, GLib.SList<E.CalComponent> old_components, GLib.SList<E.CalComponent> new_components) throws GLib.Error {
        foreach (unowned E.CalComponentId id in ids) {
            var ecal = store.get_component (id.uid, id.rid); 
            unowned iCal.Component ical = ecal.get_icalcomponent ();
            var msg = new Soup.Message ("DELETE", "https://apis.live.net/v5.0/%s".printf (ical.get_relcalid ()));
            session.send_message (msg);
            store.remove_component (id.uid, id.rid);
        }
    }

    public override void receive_objects_sync (E.DataCal cal, GLib.Cancellable? cancellable, string calobj) throws GLib.Error {
        warning ("");
    }

    public override void send_objects_sync (E.DataCal cal, GLib.Cancellable? cancellable, string calobj, GLib.SList<string> users, ref string modified_calobj) throws GLib.Error {
        warning ("");
    }

    public override void get_object_sync (E.DataCal cal, GLib.Cancellable? cancellable, string uid, string rid, ref string calobj) throws GLib.Error {
        warning ("");
    }

    public override void get_object_list_sync (E.DataCal cal, GLib.Cancellable? cancellable, string sexp, GLib.SList<string> calobjs) throws GLib.Error {
        warning ("");
    }

    public override void add_timezone_sync (E.DataCal cal, GLib.Cancellable? cancellable, string tzobject) throws GLib.Error {
        warning ("");
    }

    public override void get_free_busy_sync (E.DataCal cal, GLib.Cancellable? cancellable, GLib.SList<string> users, time_t start, time_t end, GLib.SList<string> freebusyobjects) throws GLib.Error {
        warning ("");
    }

    private void soup_authenticate (Soup.Message msg, Soup.Auth auth, bool retrying) {
        string access_token = "";
        int expires_in = 0;
        try {
            get_source ().get_oauth2_access_token_sync (null, out access_token, out expires_in);
            ((E.SoupAuthBearer) auth).set_access_token (access_token, expires_in);
        } catch (Error e) {
            critical (e.message);
        }
    }

    private void initialize_backend () throws GLib.Error {
        var source = get_source ();

        var auth_extension = (E.SourceAuthentication)source.get_extension (E.SOURCE_EXTENSION_AUTHENTICATION);
        auth_required = auth_extension.required ();

        var offline_extension = (E.SourceOffline)source.get_extension (E.SOURCE_EXTENSION_OFFLINE);
        do_offline = offline_extension.get_stay_synchronized ();

        source.changed.connect (() => {on_source_changed (source);});

        var cache_dir = get_cache_dir ();
        if (store == null) {
            store = new E.CalBackendStore (cache_dir, this);
            store.load ();
        }

        /* Set the local attachment store */
        if (GLib.DirUtils.create_with_parents (cache_dir, 0700) < 0) {
            //FIXME: throw an error, translated.
            critical ("Cannot create local cache folder '%s'", cache_dir);
            return;
        }

        if (refresh_id == 0) {
            refresh_id = source.refresh_add_timeout (null, ((src) => {
                time_to_refresh ();
            }));
        }
    }

    private void on_source_changed (E.Source source) {
        if (updating_source == true || loaded == false || is_opened () == false)
            return;

        updating_source = true;

        /* guard the call with busy_lock, thus the two threads (this 'source changed'
           thread and the 'backend open' thread) will not clash on internal data
           when they are called in once */
        initialize_backend ();

        updating_source = false;
    }

    public void time_to_refresh () {
        var ressource_extension = (E.SourceResource) get_source ().get_extension (E.SOURCE_EXTENSION_RESOURCE);
        var source_id = ressource_extension.get_identity ();
        Soup.Message msg = new Soup.Message ("GET", "https://apis.live.net/v5.0/me/calendars");
        session.send_message (msg);
        var parser = new Json.Parser ();
        try {
            parser.load_from_data ((string)msg.response_body.data);
            var elements = parser.get_root ().get_object ().dup_member ("data").get_array ().get_elements ();
            foreach (unowned Json.Node item in elements) {
                var object = item.get_object ();
                var id = object.get_string_member ("id");
                // We can't select a calendar here, let's use the default one
                if (source_id == id) {
                    var updated = object.get_string_member ("updated_time");
                    var timeval = TimeVal ();
                    timeval.from_iso8601 (updated);
                    if (last_updated == null) {
                        last_updated = new DateTime.from_timeval_local (timeval);
                        get_calendar_informations (source_id);
                    } else {
                        var new_last_updated = new DateTime.from_timeval_local (timeval);
                        if (new_last_updated.compare (last_updated) > 0) {
                            last_updated.add (last_updated.difference (new_last_updated));
                            get_calendar_informations (source_id);
                        }
                    }

                    break;
                }
            }
        } catch (Error e) {
            critical (e.message);
        }
    }

    private void get_calendar_informations (string calendar_id) {
        Soup.Message msg = new Soup.Message ("GET", "https://apis.live.net/v5.0/%s/events".printf (calendar_id));
        var now = new DateTime.now_local ();
        var start_time = now.add_years (-100).format (FORMAT);
        msg.request_headers.append ("start_time", start_time);
        var end_time = now.add_years (+100).format (FORMAT);
        msg.request_headers.append ("end_time", end_time);
        warning (msg.uri.to_string (false));
        session.send_message (msg);
        var parser = new Json.Parser ();
        try {
            warning ((string)msg.response_body.data);
            parser.load_from_data ((string)msg.response_body.data);
            var elements = parser.get_root ().get_object ().dup_member ("data").get_array ().get_elements ();
            foreach (unowned Json.Node item in elements) {
                bool found = false;
                var components = store.get_components ();
                var object = item.get_object ();
                var id = object.get_string_member ("id");
                foreach (var component in components) {
                    unowned iCal.Component ical = component.get_icalcomponent ();
                    if (ical.get_relcalid () == id) {
                        found = true;
                        update_component (component, object);
                    }
                }

                if (found == false) {
                    update_component (null, object);
                }
            }
        } catch (Error e) {
            critical (e.message);
        }
    }

    private void update_component (E.CalComponent? ecal, Json.Object object) {
        if (ecal == null) {
            var new_ecal = new E.CalComponent ();
            var id = object.get_string_member ("id");

            new_ecal.set_icalcomponent (new iCal.Component (iCal.ComponentKind.VEVENT));
            unowned iCal.Component ical = new_ecal.get_icalcomponent ();
            ical.set_relcalid (id);

            var name = object.get_string_member ("name");
            if (name != null)
                ical.set_summary (name);

            var description = object.get_string_member ("description");
            if (description != null)
                ical.set_description (description);

            var start_time = object.get_string_member ("start_time");
            if (start_time != null) {
                var timeval = TimeVal ();
                timeval.from_iso8601 (start_time);
                var start_dt = new DateTime.from_timeval_local (timeval);
                ical.set_dtstart (date_time_to_ical (start_dt, start_dt));
            }

            var end_time = object.get_string_member ("end_time");
            if (end_time != null) {
                var timeval = TimeVal ();
                timeval.from_iso8601 (end_time);
                var end_dt = new DateTime.from_timeval_local (timeval);
                ical.set_dtend (date_time_to_ical (end_dt, end_dt));
            }

            store.put_component (new_ecal);
        } else {
            unowned iCal.Component ical = ecal.get_icalcomponent ();

            var name = object.get_string_member ("name");
            if (name != null)
                ical.set_summary (name);

            var description = object.get_string_member ("description");
            if (description != null)
                ical.set_description (description);

            var start_time = object.get_string_member ("start_time");
            if (start_time != null) {
                var timeval = TimeVal ();
                timeval.from_iso8601 (start_time);
                var start_dt = new DateTime.from_timeval_local (timeval);
                ical.set_dtstart (date_time_to_ical (start_dt, start_dt));
            }

            var end_time = object.get_string_member ("end_time");
            if (end_time != null) {
                var timeval = TimeVal ();
                timeval.from_iso8601 (end_time);
                var end_dt = new DateTime.from_timeval_local (timeval);
                ical.set_dtend (date_time_to_ical (end_dt, end_dt));
            }

            var location = object.get_string_member ("location");
            if (location != null) {
                ical.set_location (location);
            }

        }
    }
}

[ModuleInit]
public static Type e_outlook_contact_fake_module_load (GLib.TypeModule type_module) {
    return typeof(E.CalBackendOutlook);
}