// -*- Mode: vala; indent-tabs-mode: nil; tab-width: 4 -*-
/*-
 * Copyright (c) 2014 Pantheon Online Accounts Developers (http://launchpad.net/pantheon-online-accounts)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authored by: Corentin Noël <corentin@elementaryos.org>
 */

// Main references : http://msdn.microsoft.com/en-us/library/hh243648.aspx
//                   http://msdn.microsoft.com/en-us/library/hh826527.aspx

public class E.BookBackendOutlook : E.BookBackend {
    private E.BookBackendCache cache;
    private Soup.Session session;
    private bool do_offline = false;
    private uint refresh_id = 0;
    List<E.DataBookView> bookviews;

    construct {
        bookviews = new List<E.DataBookView> ();
        session = new Soup.Session ();
        session.timeout = 90;
        session.ssl_strict = true;
        session.ssl_use_system_ca_file = true;
        var feature = session.get_feature (typeof (Soup.AuthManager));
        feature.add_feature (typeof (E.SoupAuthBearer));

        session.authenticate.connect ((msg, auth, retrying) => soup_authenticate (msg, auth, retrying));
    }

    public override string get_backend_property (string prop_name) {
        switch (prop_name) {
            case E.BookBackendProperty.CAPABILITIES:
                return "net,do-initial-query,contact-lists,refresh-supported";
            case E.BookBackendProperty.REQUIRED_FIELDS:
                return null; //TODO: implement it!
            case E.BookBackendProperty.SUPPORTED_FIELDS:
                return string.join (",",
                    E.Contact.field_name (E.ContactField.NAME),
                    E.Contact.field_name (E.ContactField.FULL_NAME),
                    E.Contact.field_name (E.ContactField.GIVEN_NAME),
                    E.Contact.field_name (E.ContactField.FAMILY_NAME),
                    E.Contact.field_name (E.ContactField.EMAIL_1),
                    E.Contact.field_name (E.ContactField.EMAIL_2),
                    E.Contact.field_name (E.ContactField.EMAIL_3),
                    E.Contact.field_name (E.ContactField.EMAIL_4),
                    E.Contact.field_name (E.ContactField.ANNIVERSARY));
        }

        return null;
    }

    public override void shutdown (string test) {
        
    }

    public override bool open_sync (GLib.Cancellable? cancellable = null) throws GLib.Error {
        var offline_extension = (E.SourceOffline) get_source ().get_extension (E.SOURCE_EXTENSION_OFFLINE);
        do_offline = offline_extension.get_stay_synchronized ();
        if (do_offline == false && get_online () == false) {
            throw new E.ClientError.OFFLINE_UNAVAILABLE (null);
        }

        if (cache == null) {
            var filename = Path.build_filename (get_cache_dir (), "cache.xml");
            cache = new E.BookBackendCache (filename);
            writable = false;
            set_online (true);
        }

        if (refresh_id == 0) {
            refresh_local_cache (cancellable);
            refresh_id = get_source ().refresh_add_timeout (null, (event) => {refresh_local_cache (null);});
        }

        return true;
    }

    public override bool create_contacts_sync (string[] vcards, GLib.Queue<E.Contact> out_contacts, GLib.Cancellable? cancellable = null) throws GLib.Error {
        return false;
    }

    public override bool modify_contacts_sync (string[] vcards, GLib.Cancellable? cancellable = null) throws GLib.Error {
        return false;
    }

    public override bool remove_contacts_sync (string[] uids, GLib.Cancellable? cancellable = null) throws GLib.Error {
        return false;
    }

    public override E.Contact get_contact_sync (string uid, GLib.Cancellable? cancellable = null) throws GLib.Error {
        E.Contact contact;
        lock (cache) {
            contact = cache.get_contact (uid);
        }

        return contact;
    }

    public override bool get_contact_list_sync (string query, GLib.Queue<E.Contact> out_contacts, GLib.Cancellable? cancellable = null) throws GLib.Error {
        GLib.List<E.Contact> contact_list;
        lock (cache) {
            contact_list = cache.get_contacts (query);
        }

        while (contact_list != null) {
            if (cancellable.is_cancelled () == true)
                return false;
            contact_list.remove_link (contact_list);
            out_contacts.push_tail (contact_list.data);
        }

        return true;
    }

    public override bool refresh_sync (GLib.Cancellable? cancellable = null) throws GLib.Error {
        if (get_online () == true) {
            refresh_local_cache (cancellable);
        }

        return true;
    }

    public override void start_view (E.DataBookView view) {
        var sexp = view.get_sexp ();
        var query = sexp.text ();
        GLib.List<E.Contact> contacts;
        lock (cache) {
            contacts = cache.get_contacts (query);
        }

        contacts.foreach ((contact) => {
            view.notify_update (contact);
        });

        if (get_online () == true) {
            bookviews.append (view);
            refresh_local_cache ();
        }
    }

    public override void stop_view (E.DataBookView view) {
        unowned List<E.DataBookView> list_view = bookviews.find (view);
        if (list_view != null) {
            bookviews.delete_link (list_view);
        }
    }

    private void soup_authenticate (Soup.Message msg, Soup.Auth auth, bool retrying) {
        string access_token = "";
        int expires_in = 0;
        try {
            get_source ().get_oauth2_access_token_sync (null, out access_token, out expires_in);
            ((E.SoupAuthBearer) auth).set_access_token (access_token, expires_in);
        } catch (Error e) {
            critical (e.message);
            //throw e;
        }
    }

    public void refresh_local_cache (GLib.Cancellable? cancellable = null) {
        Soup.Message msg = new Soup.Message ("GET", "https://apis.live.net/v5.0/me/contacts");
        session.send_message (msg);
        if (msg.response_body.data == null) {
            bookviews.foreach ((view) => {
                view.notify_complete (null);
            });

            return;
        }

        var parser = new Json.Parser ();
        try {
            parser.load_from_data ((string)msg.response_body.data);
            var root_object = parser.get_root ().get_object ();
            if (root_object.has_member ("error") == true) {
                
            }

            var data = root_object.dup_member ("data");
            var elements = data.dup_array ().get_elements ();
            cache.freeze_changes ();
            foreach (unowned Json.Node item in elements) {
                if (cancellable.is_cancelled () == true)
                    break;
                var object = item.get_object ();
                refresh_local_cache_from_json_object.begin (object);
            }

            var current_time = TimeVal ();
            current_time.get_current_time ();
            cache.set_time (current_time.to_iso8601 ());
            cache.thaw_changes ();
        } catch (Error e) {
            critical (e.message);
        }

        cache.set_populated ();
        bookviews.foreach ((view) => {
            view.notify_complete (null);
        });
    }

    private async void refresh_local_cache_from_json_object (Json.Object contact_object) {
        var contact = new E.Contact ();
        var updated_time = contact_object.get_string_member ("updated_time");
        var uid = contact_object.get_string_member ("id");
        if (cache.check_contact (uid) == true) {
            contact = cache.get_contact (uid);
            if (contact.Rev != null && contact.Rev == updated_time) {
                return;
            }
        }

        contact.add_attribute_with_value (new E.VCardAttribute (null, E.VC.UID), uid);
        contact.set (E.ContactField.REV, updated_time);
        contact.set (E.ContactField.FULL_NAME, contact_object.get_string_member ("name"));
        var name = new E.ContactName ();
        name.given = contact_object.get_string_member ("first_name");
        name.family = contact_object.get_string_member ("last_name");
        contact.set (E.ContactField.NAME, name);

        var emails_object = contact_object.get_object_member ("emails");
        var primary_email = emails_object.get_string_member ("preferred");
        if (primary_email != null) {
            primary_email = "";
        }

        var email_string = emails_object.get_string_member ("account");
        if (email_string != null) {
            var attribute = new E.VCardAttribute (null, E.VC.EMAIL);
            attribute.add_value (email_string);
            if (email_string == primary_email) {
                attribute.add_param_with_value (new E.VCardAttributeParam ("type"), "pref");
            }

            contact.add_attribute (attribute);
        }

        email_string = emails_object.get_string_member ("personal");
        if (email_string != null) {
            var attribute = new E.VCardAttribute (null, E.VC.EMAIL);
            attribute.add_value (email_string);
            attribute.add_param_with_value (new E.VCardAttributeParam ("type"), "HOME");
            if (email_string == primary_email) {
                attribute.add_param_with_value (new E.VCardAttributeParam ("type"), "pref");
            }

            contact.add_attribute (attribute);
        }

        email_string = emails_object.get_string_member ("business");
        if (email_string != null) {
            var attribute = new E.VCardAttribute (null, E.VC.EMAIL);
            attribute.add_value (email_string);
            attribute.add_param_with_value (new E.VCardAttributeParam ("type"), "WORK");
            if (email_string == primary_email) {
                attribute.add_param_with_value (new E.VCardAttributeParam ("type"), "pref");
            }

            contact.add_attribute (attribute);
        }

        email_string = emails_object.get_string_member ("other");
        if (email_string != null) {
            var attribute = new E.VCardAttribute (null, E.VC.EMAIL);
            attribute.add_param_with_value (new E.VCardAttributeParam ("type"), "OTHER");
            attribute.add_value (email_string);
            if (email_string == primary_email) {
                attribute.add_param_with_value (new E.VCardAttributeParam ("type"), "pref");
            }

            contact.add_attribute (attribute);
        }

        if (contact_object.get_null_member ("birth_day") == false && contact_object.get_null_member ("birth_month") == false) {
            var contact_date = new E.ContactDate ();
            contact_date.day = (uint)contact_object.get_int_member ("birth_day");
            contact_date.month = (uint)contact_object.get_int_member ("birth_month");
            contact.set (E.ContactField.BIRTH_DATE, contact_date);
        }

        if (contact_object.get_null_member ("user_id") == false) {
            var user_id = contact_object.get_string_member ("user_id");
            set_contact_picture_from_user_id.begin (contact, user_id);
            
        }

        lock (cache) {
            cache.add_contact (contact);
        }

        bookviews.foreach ((view) => {
            view.notify_update (contact);
        });
    }

    private async void set_contact_picture_from_user_id (E.Contact contact, string user_id) {
        Soup.Message msg = new Soup.Message ("GET", "https://apis.live.net/v5.0/%s/picture".printf (user_id));
        session.send_message (msg);
        var contact_photo = new E.ContactPhoto ();
        contact_photo.set_inlined (msg.response_body.data);
        contact.set (E.ContactField.PHOTO, contact_photo);
    }
}

[ModuleInit]
public static Type e_outlook_addressbook_fake_module_load (GLib.TypeModule type_module) {
    return typeof(E.BookBackendOutlook);
}